@@ -12,6 +12,7 @@
 #include <stdlib.h>
 #include <cassert>
 #include <string>
+#include <vector>
 
 #include "xmlrpc-c/girerr.hpp"
 using girerr::error;
@@ -19,12 +20,14 @@ using girerr::throwf;
 #include "xmlrpc-c/girmem.hpp"
 using girmem::autoObjectPtr;
 using girmem::autoObject;
+#include "env_wrap.hpp"
 #include "xmlrpc-c/base.h"
 #include "xmlrpc-c/client.h"
 #include "xmlrpc-c/transport.h"
 #include "xmlrpc-c/base.hpp"
 #include "xmlrpc-c/xml.hpp"
 #include "xmlrpc-c/client.hpp"
+#include "transport_config.h"
 
 using namespace std;
 using namespace xmlrpc_c;
@@ -32,23 +35,28 @@ using namespace xmlrpc_c;
 
 namespace {
 
+void
+throwIfError(env_wrap const& env) {
+
+    if (env.env_c.fault_occurred)
+        throw(error(env.env_c.fault_string));
+}
+
+
+
 class memblockStringWrapper {
 
 public:    
     memblockStringWrapper(string const value) {
 
-        xmlrpc_env env;
-        xmlrpc_env_init(&env);
+        env_wrap env;
 
-        this->memblockP = XMLRPC_MEMBLOCK_NEW(char, &env, 0);
+        this->memblockP = XMLRPC_MEMBLOCK_NEW(char, &env.env_c, 0);
+        throwIfError(env);
 
-        if (env.fault_occurred)
-            throw(error(env.fault_string));
-
-        XMLRPC_MEMBLOCK_APPEND(char, &env, this->memblockP,
+        XMLRPC_MEMBLOCK_APPEND(char, &env.env_c, this->memblockP,
                                value.c_str(), value.size());
-        if (env.fault_occurred)
-            throw(error(env.fault_string));
+        throwIfError(env);
     }
     
     memblockStringWrapper(xmlrpc_mem_block * const memblockP) :
@@ -96,7 +104,7 @@ carriageParmPtr::operator->() const {
 
 
 carriageParm *
-carriageParmPtr::getp() const {
+carriageParmPtr::get() const {
     return dynamic_cast<carriageParm *>(objectP);
 }
 
@@ -129,13 +137,11 @@ carriageParm_http0::instantiate(string const serverUrl) {
     if (c_serverInfoP)
         throw(error("object already instantiated"));
     
-    xmlrpc_env env;
-    xmlrpc_env_init(&env);
+    env_wrap env;
 
-    this->c_serverInfoP = xmlrpc_server_info_new(&env, serverUrl.c_str());
-
-    if (env.fault_occurred)
-        throw(error(env.fault_string));
+    this->c_serverInfoP =
+        xmlrpc_server_info_new(&env.env_c, serverUrl.c_str());
+    throwIfError(env);
 }
 
 
@@ -147,14 +153,11 @@ carriageParm_http0::setBasicAuth(string const username,
     if (!c_serverInfoP)
         throw(error("object not instantiated"));
     
-    xmlrpc_env env;
-    xmlrpc_env_init(&env);
+    env_wrap env;
 
     xmlrpc_server_info_set_basic_auth(
-        &env, this->c_serverInfoP, username.c_str(), password.c_str());
-
-    if (env.fault_occurred)
-        throw(error(env.fault_string));
+        &env.env_c, this->c_serverInfoP, username.c_str(), password.c_str());
+    throwIfError(env);
 }
 
 
@@ -288,7 +291,7 @@ clientXmlTransportPtr::clientXmlTransportPtr(
 
 
 clientXmlTransport *
-clientXmlTransportPtr::getp() const {
+clientXmlTransportPtr::get() const {
     return dynamic_cast<clientXmlTransport *>(objectP);
 }
 
@@ -313,9 +316,6 @@ clientXmlTransport_http::call(
     string         const& callXml,
     string *       const  responseXmlP) {
 
-    xmlrpc_env env;
-    xmlrpc_env_init(&env);
-
     carriageParm_http0 * const carriageParmHttpP =
         dynamic_cast<carriageParm_http0 *>(carriageParmP);
 
@@ -327,14 +327,15 @@ clientXmlTransport_http::call(
 
     xmlrpc_mem_block * responseXmlMP;
 
-    this->c_transportOpsP->call(&env,
+    env_wrap env;
+
+    this->c_transportOpsP->call(&env.env_c,
                                 this->c_transportP,
                                 carriageParmHttpP->c_serverInfoP,
                                 callXmlM.memblockP,
                                 &responseXmlMP);
 
-    if (env.fault_occurred)
-        throw(error(env.fault_string));
+    throwIfError(env);
 
     memblockStringWrapper responseHolder(responseXmlMP);
         // Makes responseXmlMP get freed at end of scope
@@ -351,8 +352,7 @@ clientXmlTransport_http::start(
     string            const& callXml,
     xmlTransactionPtr const& xmlTranP) {
 
-    xmlrpc_env env;
-    xmlrpc_env_init(&env);
+    env_wrap env;
 
     carriageParm_http0 * const carriageParmHttpP =
         dynamic_cast<carriageParm_http0 *>(carriageParmP);
@@ -371,15 +371,14 @@ clientXmlTransport_http::start(
 
     try {
         this->c_transportOpsP->send_request(
-            &env,
+            &env.env_c,
             this->c_transportP,
             carriageParmHttpP->c_serverInfoP,
             callXmlM.memblockP,
             &this->asyncComplete,
             reinterpret_cast<xmlrpc_call_info *>(xmlTranP2P));
 
-        if (env.fault_occurred)
-            throw(error(env.fault_string));
+        throwIfError(env);
     } catch (...) {
         delete xmlTranP2P;
         throw;
@@ -400,6 +399,71 @@ clientXmlTransport_http::finishAsync(xmlrpc_c::timeout const timeout) {
 }
 
 
+bool const haveCurl(
+#if MUST_BUILD_CURL_CLIENT
+true
+#else
+false
+#endif
+);
+
+bool const haveLibwww(
+#if MUST_BUILD_LIBWWW_CLIENT
+true
+#else
+false
+#endif
+);
+
+bool const haveWininet(
+#if MUST_BUILD_WININET_CLIENT
+true
+#else
+false
+#endif
+);
+
+
+
+vector<string>
+clientXmlTransport_http::availableTypes() {
+
+    vector<string> retval;
+
+    if (haveCurl)
+        retval.push_back("curl");
+
+    if (haveLibwww)
+        retval.push_back("libwww");
+
+    if (haveWininet)
+        retval.push_back("wininet");
+
+    return retval;
+}
+
+
+
+clientXmlTransportPtr
+clientXmlTransport_http::create() {
+/*----------------------------------------------------------------------------
+  Make an HTTP Client XML transport of any kind (Caller doesn't care).
+
+  Caller can find out what kind he got by trying dynamic casts.
+
+  Caller can use a carriageParm_http0 with the transport.
+-----------------------------------------------------------------------------*/
+    if (haveCurl)
+        return clientXmlTransportPtr(new clientXmlTransport_curl());
+    else if (haveLibwww)
+        return clientXmlTransportPtr(new clientXmlTransport_libwww());
+    else if (haveWininet)
+        return clientXmlTransportPtr(new clientXmlTransport_wininet());
+    else
+        throwf("This XML-RPC client library contains no HTTP XML transports");
+}
+
+
 
 clientTransaction::clientTransaction() {}
 
@@ -424,9 +488,10 @@ clientTransactionPtr::operator->() const {
 client::~client() {}
 
 
+
 void
 client::start(carriageParm *       const  carriageParmP,
-              string               const  methodName,
+              string               const& methodName,
               paramList            const& paramList,
               clientTransactionPtr const& tranP) {
 /*----------------------------------------------------------------------------
@@ -467,6 +532,13 @@ clientPtr::operator->() const {
 
 
 
+client *
+clientPtr::get() const {
+    return dynamic_cast<client *>(objectP);
+}
+
+
+
 client_xml::client_xml(clientXmlTransport * const transportP) :
     transportP(transportP) {}
 
@@ -475,20 +547,14 @@ client_xml::client_xml(clientXmlTransport * const transportP) :
 client_xml::client_xml(clientXmlTransportPtr const transportPtr) {
 
     this->transportPtr = transportPtr;
-    try {
-        this->transportP = transportPtr.getp();
-    } catch (...) {
-        /* Constructor is failing; destroy what we've built so far. */
-        this->transportPtr.unpoint();
-        throw;
-    }
+    this->transportP   = transportPtr.get();
 }
      
 
 
 void
 client_xml::call(carriageParm * const  carriageParmP,
-                 string         const  methodName,
+                 string         const& methodName,
                  paramList      const& paramList,
                  rpcOutcome *   const  outcomeP) {
 
@@ -519,7 +585,7 @@ client_xml::call(carriageParm * const  carriageParmP,
 
 void
 client_xml::start(carriageParm *       const  carriageParmP,
-                  string               const  methodName,
+                  string               const& methodName,
                   paramList            const& paramList,
                   clientTransactionPtr const& tranP) {
 
@@ -554,9 +620,9 @@ serverAccessor::serverAccessor(clientPtr       const clientP,
 void
 serverAccessor::call(std::string            const& methodName,
                      xmlrpc_c::paramList    const& paramList,
-                     xmlrpc_c::rpcOutcome * const  outcomeP) {
+                     xmlrpc_c::rpcOutcome * const  outcomeP) const {
 
-    this->clientP->call(this->carriageParmP.getp(),
+    this->clientP->call(this->carriageParmP.get(),
                         methodName,
                         paramList,
                         outcomeP);
@@ -564,6 +630,35 @@ serverAccessor::call(std::string            const& methodName,
 
 
 
+serverAccessorPtr::serverAccessorPtr() {
+    // Base class constructor will construct pointer that points to nothing
+}
+
+
+
+serverAccessorPtr::serverAccessorPtr(
+    serverAccessor * const serverAccessorParmP) {
+    this->point(serverAccessorParmP);
+}
+
+
+
+serverAccessor *
+serverAccessorPtr::operator->() const {
+
+    autoObject * const p(this->objectP);
+    return dynamic_cast<serverAccessor *>(p);
+}
+
+
+
+serverAccessor *
+serverAccessorPtr::get() const {
+    return dynamic_cast<serverAccessor *>(objectP);
+}
+
+
+
 connection::connection(client *       const clientP,
                        carriageParm * const carriageParmP) :
     clientP(clientP), carriageParmP(carriageParmP) {}
